(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('@angular/core'), require('@angular/common'), require('@angular/forms')) :
	typeof define === 'function' && define.amd ? define(['exports', '@angular/core', '@angular/common', '@angular/forms'], factory) :
	(factory((global['angular-formatter-parser'] = global['angular-formatter-parser'] || {}),global._angular_core,global._angular_common,global._angular_forms));
}(this, (function (exports,_angular_core,_angular_common,_angular_forms) { 'use strict';

var FORMATTER_PARSER = new _angular_core.InjectionToken('formatterParser');

var emptyString = '';
var emptyArray = [];
var defaultPlaceholderChar = '_';
var convertMaskToPlaceholder = function (mask, placeholderChar) {
    mask = mask || emptyArray;
    placeholderChar = placeholderChar || defaultPlaceholderChar;
    if (mask.indexOf(placeholderChar) !== -1) {
        throw new Error('Placeholder character must not be used as part of the mask. Please specify a character ' +
            'that is not present in your mask as your placeholder character.\n\n' +
            ("The placeholder character that was received is: " + JSON.stringify(placeholderChar) + "\n\n") +
            ("The mask that was received is: " + JSON.stringify(mask)));
    }
    return mask.map(function (char) {
        return (char instanceof RegExp) ? placeholderChar : char;
    }).join('');
};
var FormatterParser = (function () {
    function FormatterParser() {
    }
    /**
     * @param {?} searchValue
     * @param {?} replaceValue
     * @return {?}
     */
    FormatterParser.replaceString = function (searchValue, replaceValue) {
        return function (value) {
            var /** @type {?} */ transformedValue = value;
            if (typeof transformedValue === 'string' || transformedValue instanceof String) {
                transformedValue = transformedValue.replace(searchValue, replaceValue);
            }
            var /** @type {?} */ result = {
                name: 'replaceString',
                result: transformedValue,
                previous: value
            };
            return result;
        };
    };
    /**
     * @param {?=} mask
     * @param {?=} config
     * @return {?}
     */
    FormatterParser.conformToMask = function (mask, config) {
        if (mask === void 0) { mask = emptyArray; }
        config = config || ({});
        // These configurations tell us how to conform the mask
        var /** @type {?} */ guide = config.guide;
        var /** @type {?} */ placeholderChar = config.placeholderChar || defaultPlaceholderChar;
        var /** @type {?} */ currentCaretPosition = config.currentCaretPosition;
        var /** @type {?} */ keepCharPositions = config.keepCharPositions;
        // these are internal states
        var /** @type {?} */ previousConformedValue = '';
        return function (rawValue) {
            rawValue = rawValue || '';
            if (typeof mask === 'function') {
                mask = (mask(rawValue));
            }
            var /** @type {?} */ placeholder = convertMaskToPlaceholder(mask, placeholderChar);
            // The configs below indicate that the user wants the algorithm to work in *no guide* mode
            var /** @type {?} */ suppressGuide = guide === false && previousConformedValue !== undefined;
            // Calculate lengths once for performance
            var /** @type {?} */ rawValueLength = rawValue.length;
            var /** @type {?} */ previousConformedValueLength = previousConformedValue.length;
            var /** @type {?} */ placeholderLength = placeholder.length;
            var /** @type {?} */ maskLength = mask.length;
            // This tells us the number of edited characters and the direction in which they were edited (+/-)
            var /** @type {?} */ editDistance = rawValueLength - previousConformedValueLength;
            // In *no guide* mode, we need to know if the user is trying to add a character or not
            var /** @type {?} */ isAddition = editDistance > 0;
            // Tells us the index of the first change. For (438) 394-4938 to (38) 394-4938, that would be 1
            var /** @type {?} */ indexOfFirstChange = currentCaretPosition + (isAddition ? -editDistance : 0);
            // We're also gonna need the index of last change, which we can derive as follows...
            var /** @type {?} */ indexOfLastChange = indexOfFirstChange + Math.abs(editDistance);
            // If `conformToMask` is configured to keep character positions, that is, for mask 111, previous value
            // _2_ and raw value 3_2_, the new conformed value should be 32_, not 3_2 (default behavior). That's in the case of
            // addition. And in the case of deletion, previous value _23, raw value _3, the new conformed string should be
            // __3, not _3_ (default behavior)
            //
            // The next block of logic handles keeping character positions for the case of deletion. (Keeping
            // character positions for the case of addition is further down since it is handled differently.)
            // To do this, we want to compensate for all characters that were deleted
            if (keepCharPositions === true && !isAddition) {
                // We will be storing the new placeholder characters in this variable.
                var /** @type {?} */ compensatingPlaceholderChars = emptyString;
                // For every character that was deleted from a placeholder position, we add a placeholder char
                for (var /** @type {?} */ i = indexOfFirstChange; i < indexOfLastChange; i++) {
                    if (placeholder[i] === placeholderChar) {
                        compensatingPlaceholderChars += placeholderChar;
                    }
                }
                // Now we trick our algorithm by modifying the raw value to make it contain additional placeholder characters
                // That way when the we start laying the characters again on the mask, it will keep the non-deleted characters
                // in their positions.
                rawValue = (rawValue.slice(0, indexOfFirstChange) +
                    compensatingPlaceholderChars +
                    rawValue.slice(indexOfFirstChange, rawValueLength));
            }
            // Convert `rawValue` string to an array, and mark characters based on whether they are newly added or have
            // existed in the previous conformed value. Identifying new and old characters is needed for `conformToMask`
            // to work if it is configured to keep character positions.
            var /** @type {?} */ rawValueArr = rawValue
                .split(emptyString)
                .map(function (char, i) { return ({
                char: char,
                isNew: i >= indexOfFirstChange && i < indexOfLastChange
            }); });
            // The loop below removes masking characters from user input. For example, for mask
            // `00 (111)`, the placeholder would be `00 (___)`. If user input is `00 (234)`, the loop below
            // would remove all characters but `234` from the `rawValueArr`. The rest of the algorithm
            // then would lay `234` on top of the available placeholder positions in the mask.
            for (var /** @type {?} */ i = rawValueLength - 1; i >= 0; i--) {
                var char = rawValueArr[i].char;
                if (char !== placeholderChar) {
                    var /** @type {?} */ shouldOffset = i >= indexOfFirstChange && previousConformedValueLength === maskLength;
                    if (char === placeholder[(shouldOffset) ? i - editDistance : i]) {
                        rawValueArr.splice(i, 1);
                    }
                }
            }
            // This is the variable that we will be filling with characters as we figure them out
            // in the algorithm below
            var /** @type {?} */ conformedValue = emptyString;
            var /** @type {?} */ someCharsRejected = false;
            // Ok, so first we loop through the placeholder looking for placeholder characters to fill up.
            placeholderLoop: for (var /** @type {?} */ i = 0; i < placeholderLength; i++) {
                var /** @type {?} */ charInPlaceholder = placeholder[i];
                // We see one. Let's find out what we can put in it.
                if (charInPlaceholder === placeholderChar) {
                    // But before that, do we actually have any user characters that need a place?
                    if (rawValueArr.length > 0) {
                        // We will keep chipping away at user input until either we run out of characters
                        // or we find at least one character that we can map.
                        while (rawValueArr.length > 0) {
                            // Let's retrieve the first user character in the queue of characters we have left
                            var _a = rawValueArr.shift(), rawValueChar = _a.char, isNew = _a.isNew;
                            // If the character we got from the user input is a placeholder character (which happens
                            // regularly because user input could be something like (540) 90_-____, which includes
                            // a bunch of `_` which are placeholder characters) and we are not in *no guide* mode,
                            // then we map this placeholder character to the current spot in the placeholder
                            if (rawValueChar === placeholderChar && suppressGuide !== true) {
                                conformedValue += placeholderChar;
                                // And we go to find the next placeholder character that needs filling
                                continue placeholderLoop;
                                // Else if, the character we got from the user input is not a placeholder, let's see
                                // if the current position in the mask can accept it.
                            }
                            else if (((mask[i])).test(rawValueChar)) {
                                // we map the character differently based on whether we are keeping character positions or not.
                                // If any of the conditions below are met, we simply map the raw value character to the
                                // placeholder position.
                                if (keepCharPositions !== true ||
                                    isNew === false ||
                                    previousConformedValue === emptyString ||
                                    guide === false ||
                                    !isAddition) {
                                    conformedValue += rawValueChar;
                                }
                                else {
                                    // We enter this block of code if we are trying to keep character positions and none of the conditions
                                    // above is met. In this case, we need to see if there's an available spot for the raw value character
                                    // to be mapped to. If we couldn't find a spot, we will discard the character.
                                    //
                                    // For example, for mask `1111`, previous conformed value `_2__`, raw value `942_2__`. We can map the
                                    // `9`, to the first available placeholder position, but then, there are no more spots available for the
                                    // `4` and `2`. So, we discard them and end up with a conformed value of `92__`.
                                    var /** @type {?} */ rawValueArrLength = rawValueArr.length;
                                    var /** @type {?} */ indexOfNextAvailablePlaceholderChar = null;
                                    // Let's loop through the remaining raw value characters. We are looking for either a suitable spot, ie,
                                    // a placeholder character or a non-suitable spot, ie, a non-placeholder character that is not new.
                                    // If we see a suitable spot first, we store its position and exit the loop. If we see a non-suitable
                                    // spot first, we exit the loop and our `indexOfNextAvailablePlaceholderChar` will stay as `null`.
                                    for (var /** @type {?} */ j = 0; j < rawValueArrLength; j++) {
                                        var /** @type {?} */ charData = rawValueArr[j];
                                        if (charData.char !== placeholderChar && charData.isNew === false) {
                                            break;
                                        }
                                        if (charData.char === placeholderChar) {
                                            indexOfNextAvailablePlaceholderChar = j;
                                            break;
                                        }
                                    }
                                    // If `indexOfNextAvailablePlaceholderChar` is not `null`, that means the character is not blocked.
                                    // We can map it. And to keep the character positions, we remove the placeholder character
                                    // from the remaining characters
                                    if (indexOfNextAvailablePlaceholderChar !== null) {
                                        conformedValue += rawValueChar;
                                        rawValueArr.splice(indexOfNextAvailablePlaceholderChar, 1);
                                        // If `indexOfNextAvailablePlaceholderChar` is `null`, that means the character is blocked. We have to
                                        // discard it.
                                    }
                                    else {
                                        i--;
                                    }
                                }
                                // Since we've mapped this placeholder position. We move on to the next one.
                                continue placeholderLoop;
                            }
                            else {
                                someCharsRejected = true;
                            }
                        }
                    }
                    // We reach this point when we've mapped all the user input characters to placeholder
                    // positions in the mask. In *guide* mode, we append the left over characters in the
                    // placeholder to the `conformedString`, but in *no guide* mode, we don't wanna do that.
                    //
                    // That is, for mask `(111)` and user input `2`, we want to return `(2`, not `(2__)`.
                    if (suppressGuide === false) {
                        conformedValue += placeholder.substr(i, placeholderLength);
                    }
                    // And we break
                    break;
                    // Else, the charInPlaceholder is not a placeholderChar. That is, we cannot fill it
                    // with user input. So we just map it to the final output
                }
                else {
                    conformedValue += charInPlaceholder;
                }
            }
            // The following logic is needed to deal with the case of deletion in *no guide* mode.
            //
            // Consider the silly mask `(111) /// 1`. What if user tries to delete the last placeholder
            // position? Something like `(589) /// `. We want to conform that to `(589`. Not `(589) /// `.
            // That's why the logic below finds the last filled placeholder character, and removes everything
            // from that point on.
            if (suppressGuide && isAddition === false) {
                var /** @type {?} */ indexOfLastFilledPlaceholderChar = null;
                // Find the last filled placeholder position and substring from there
                for (var /** @type {?} */ i = 0; i < conformedValue.length; i++) {
                    if (placeholder[i] === placeholderChar) {
                        indexOfLastFilledPlaceholderChar = i;
                    }
                }
                if (indexOfLastFilledPlaceholderChar !== null) {
                    // We substring from the beginning until the position after the last filled placeholder char.
                    conformedValue = conformedValue.substr(0, indexOfLastFilledPlaceholderChar + 1);
                }
                else {
                    // If we couldn't find `indexOfLastFilledPlaceholderChar` that means the user deleted
                    // the first character in the mask. So we return an empty string.
                    conformedValue = emptyString;
                }
            }
            previousConformedValue = conformedValue;
            return {
                name: 'conformToMask',
                result: conformedValue,
                previous: rawValue,
                meta: { someCharsRejected: someCharsRejected }
            };
        };
    };
    return FormatterParser;
}());
FormatterParser.toUpperCase = function (value) {
    var /** @type {?} */ transformedValue = value;
    if (typeof value === 'string' || value instanceof String) {
        transformedValue = value.toUpperCase();
    }
    return {
        name: 'toUpperCase',
        result: transformedValue,
        previous: value
    };
};
FormatterParser.toLowerCase = function (value) {
    var /** @type {?} */ transformedValue = value;
    if (typeof transformedValue === 'string' || transformedValue instanceof String) {
        transformedValue = transformedValue.toLowerCase();
    }
    return {
        name: 'toLowerCase',
        result: transformedValue,
        previous: value
    };
};
FormatterParser.toCapitalized = function (value) {
    var /** @type {?} */ transformedValue = value;
    if (typeof value === 'string' || value instanceof String) {
        transformedValue = transformedValue
            .toLowerCase()
            .split(' ')
            .map(function (val) { return val.charAt(0).toUpperCase() + val.slice(1); })
            .join(' ');
    }
    return {
        name: 'toCapitalized',
        result: transformedValue,
        previous: value
    };
};

var FormatterParserService = (function () {
    /**
     * @param {?} FORMATTER_PARSER
     */
    function FormatterParserService(FORMATTER_PARSER$$1) {
        this.FORMATTER_PARSER = FORMATTER_PARSER$$1;
    }
    /**
     * @param {?} functionName
     * @param {?=} params
     * @return {?}
     */
    FormatterParserService.prototype.getFormatParseFunction = function (functionName, params) {
        var /** @type {?} */ formatParseFunction;
        if (functionName in FormatterParser) {
            formatParseFunction = FormatterParser[functionName];
        }
        else if (this.FORMATTER_PARSER) {
            formatParseFunction = this.FORMATTER_PARSER.find(function (formParsFunc) {
                return functionName === formParsFunc.name;
            });
        }
        if (!(typeof formatParseFunction === 'function')) {
            throw new Error("Formatter or Parser with name " + functionName + " \n                          is not provided as a function via FormatterParser \n                          service or FORMATTER_PARSER InjectionToken. \n                          Did you forgot to provide them?");
        }
        return (params) ? formatParseFunction.apply(void 0, params) : formatParseFunction;
    };
    return FormatterParserService;
}());
FormatterParserService.decorators = [
    { type: _angular_core.Injectable },
];
/**
 * @nocollapse
 */
FormatterParserService.ctorParameters = function () { return [
    { type: Array, decorators: [{ type: _angular_core.Optional }, { type: _angular_core.Inject, args: [FORMATTER_PARSER,] },] },
]; };

var CONTROL_VALUE_ACCESSOR = {
    name: 'formatterParserValueAccessor',
    provide: _angular_forms.NG_VALUE_ACCESSOR,
    useExisting: _angular_core.forwardRef(function () { return FormatterParserDirective; }),
    multi: true
};
var FormatterParserDirective = (function () {
    /**
     * @param {?} _elementRef
     * @param {?} fps
     * @param {?} fcd
     */
    function FormatterParserDirective(_elementRef, fps, fcd) {
        this._elementRef = _elementRef;
        this.fps = fps;
        this.fcd = fcd;
        this.formatterParserView = [];
        this.formatterParserModel = [];
    }
    /**
     * @param {?} fn
     * @return {?}
     */
    FormatterParserDirective.prototype.registerOnTouched = function (fn) {
        this.onTouch = fn;
    };
    /**
     * @param {?} fn
     * @return {?}
     */
    FormatterParserDirective.prototype.registerOnChange = function (fn) {
        this.onModelChange = fn;
    };
    /**
     * @return {?}
     */
    FormatterParserDirective.prototype.ngOnInit = function () {
        this.formControl = ((this.fcd)).form.controls[this.formControlName];
        this.inputElement = this.getInputElementRef();
        this.updateFormatterAndParser();
    };
    /**
     * @param {?} $event
     * @return {?}
     */
    FormatterParserDirective.prototype.onControlInput = function ($event) {
        var /** @type {?} */ rawValue = this.inputElement.value;
        // If there is a reactive FormControl present trigger onTouch
        if (this.onTouch) {
            this.onTouch();
        }
        // write value to view (visible text of the form control)
        this.inputElement.value = this.formatterParserView
            .reduce(function (state, transform) { return transform(state).result; }, rawValue || null);
        // write value to model (value stored in FormControl)
        var /** @type {?} */ modelValue = this.formatterParserModel
            .reduce(function (state, transform) { return transform(state).result; }, rawValue || null);
        // If there is a reactive formControl present update its model
        if (this.onModelChange) {
            this.onModelChange(modelValue);
        }
    };
    /**
     * @param {?} rawValue
     * @return {?}
     */
    FormatterParserDirective.prototype.writeValue = function (rawValue) {
        // write value to view (visible text of the form control)
        this.inputElement.value = this.formatterParserView
            .reduce(function (state, transform) { return transform(state).result; }, rawValue);
        // write value to model (value stored in FormControl)
        var /** @type {?} */ modelValue = this.formatterParserModel
            .reduce(function (state, transform) { return transform(state).result; }, rawValue);
        // prevent cyclic function calls
        if (rawValue !== modelValue) {
            // If there is a reactive FormControl present update its model
            if (this.onModelChange) {
                // @TODO consider other way to call patchValue
                this.formControl.setValue(modelValue, { emitViewToModelChange: true });
            }
        }
    };
    /**
     * @return {?}
     */
    FormatterParserDirective.prototype.updateFormatterAndParser = function () {
        var _this = this;
        this.formatterParserView = [];
        this.formatterParserModel = [];
        if (!this.config) {
            return;
        }
        if ('formatterParser' in this.config) {
            // setup formatterParser functions for view and model values
            this.config.formatterParser
                .forEach(function (formatterConfig) {
                var /** @type {?} */ targetBoth = 2;
                var /** @type {?} */ fPF = _this.fps.getFormatParseFunction(formatterConfig.name, formatterConfig.params);
                var /** @type {?} */ t = (formatterConfig.target === undefined) ? targetBoth : formatterConfig.target;
                if (t === 1 || t === 2) {
                    _this.formatterParserModel.push(fPF);
                }
                if ((t === 0 || t === 2)) {
                    _this.formatterParserView.push(fPF);
                }
            });
        }
    };
    /**
     * @return {?}
     */
    FormatterParserDirective.prototype.getInputElementRef = function () {
        var /** @type {?} */ input;
        if (this._elementRef.nativeElement.tagName === 'INPUT') {
            // `textMask` directive is used directly on an input element
            input = this._elementRef.nativeElement;
        }
        else {
            // `textMask` directive is used on an abstracted input element, `ion-input`, `md-input`, etc
            input = this._elementRef.nativeElement.getElementsByTagName('INPUT')[0];
        }
        if (!input) {
            throw new Error('You can applied the "formatterParser" directive only on inputs or elements containing inputs');
        }
        return input;
    };
    return FormatterParserDirective;
}());
FormatterParserDirective.decorators = [
    { type: _angular_core.Directive, args: [{
                selector: '[formatterParser]',
                providers: [
                    CONTROL_VALUE_ACCESSOR
                ]
            },] },
];
/**
 * @nocollapse
 */
FormatterParserDirective.ctorParameters = function () { return [
    { type: _angular_core.ElementRef, },
    { type: FormatterParserService, },
    { type: _angular_forms.ControlContainer, decorators: [{ type: _angular_core.Optional }, { type: _angular_core.Host }, { type: _angular_core.SkipSelf },] },
]; };
FormatterParserDirective.propDecorators = {
    'config': [{ type: _angular_core.Input, args: ['formatterParser',] },],
    'formControlName': [{ type: _angular_core.Input },],
    'onControlInput': [{ type: _angular_core.HostListener, args: ['input', ['$event'],] },],
};

var FormatterParserModule = (function () {
    function FormatterParserModule() {
    }
    /**
     * @return {?}
     */
    FormatterParserModule.forRoot = function () {
        return {
            ngModule: FormatterParserModule,
            providers: [
                FormatterParserService,
                { provide: FormatterParser, useClass: FormatterParser }
            ]
        };
    };
    return FormatterParserModule;
}());
FormatterParserModule.decorators = [
    { type: _angular_core.NgModule, args: [{
                imports: [
                    _angular_common.CommonModule,
                    _angular_forms.ReactiveFormsModule
                ],
                declarations: [FormatterParserDirective],
                exports: [FormatterParserDirective, _angular_forms.ReactiveFormsModule]
            },] },
];
/**
 * @nocollapse
 */
FormatterParserModule.ctorParameters = function () { return []; };

exports.FormatterParserModule = FormatterParserModule;
exports.FormatterParserDirective = FormatterParserDirective;
exports.FormatterParserService = FormatterParserService;
exports.FormatterParser = FormatterParser;
exports.FORMATTER_PARSER = FORMATTER_PARSER;

Object.defineProperty(exports, '__esModule', { value: true });

})));
